#' CLAWS: CLustering with wAvelets and Wer distanceS
#'
#' Groups electricity power curves (or any series of similar nature) by applying PAM
#' algorithm in parallel to chunks of size \code{nb_series_per_chunk}. Input series
#' must be sampled on the same time grid, no missing values.
#'
#' @param getSeries Access to the (time-)series, which can be of one of the three
#'   following types:
#'   \itemize{
#'     \item matrix: each line contains all the values for one time-serie, ordered by time
#'     \item connection: any R connection object (e.g. a file) providing lines as described above
#'     \item function: a custom way to retrieve the curves; it has only one argument:
#'       the indices of the series to be retrieved. See examples
#'   }
#' @inheritParams clustering
#' @param K1 Number of super-consumers to be found after stage 1 (K1 << N)
#' @param K2 Number of clusters to be found after stage 2 (K2 << K1)
#' @param wf Wavelet transform filter; see ?wavelets::wt.filter
#' @param ctype Type of contribution: "relative" or "absolute" (or any prefix)
#' @param WER "end" to apply stage 2 after stage 1 has fully iterated, or "mix" to apply stage 2
#'   at the end of each task
#' @param random TRUE (default) for random chunks repartition
#' @param ntasks Number of tasks (parallel iterations to obtain K1 medoids); default: 1.
#'   Note: ntasks << N, so that N is "roughly divisible" by N (number of series)
#' @param ncores_tasks "MPI" number of parallel tasks (1 to disable: sequential tasks)
#' @param ncores_clust "OpenMP" number of parallel clusterings in one task
#' @param nb_series_per_chunk (~Maximum) number of series in each group, inside a task
#' @param min_series_per_chunk Minimum number of series in each group
#' @param sep Separator in CSV input file (if any provided)
#' @param nbytes Number of bytes to serialize a floating-point number; 4 or 8
#' @param endian Endianness to use for (de)serialization. Use "little" or "big" for portability
#' @param verbose Level of verbosity (0/FALSE for nothing or 1/TRUE for all; devel stage)
#'
#' @return A matrix of the final medoids curves (K2) in rows
#'
#' @examples
#' \dontrun{
#' # WER distances computations are a bit too long for CRAN (for now)
#'
#' # Random series around cos(x,2x,3x)/sin(x,2x,3x)
#' x = seq(0,500,0.05)
#' L = length(x) #10001
#' ref_series = matrix( c(cos(x), cos(2*x), cos(3*x), sin(x), sin(2*x), sin(3*x)),
#'   byrow=TRUE, ncol=L )
#' library(wmtsa)
#' series = do.call( rbind, lapply( 1:6, function(i)
#'   do.call(rbind, wmtsa::wavBootstrap(ref_series[i,], n.realization=400)) ) )
#' #dim(series) #c(2400,10001)
#' medoids_ascii = claws(series, K1=60, K2=6, "d8", "rel", nb_series_per_chunk=500)
#'
#' # Same example, from CSV file
#' csv_file = "/tmp/epclust_series.csv"
#' write.table(series, csv_file, sep=",", row.names=FALSE, col.names=FALSE)
#' medoids_csv = claws(csv_file, K1=60, K2=6, "d8", "rel", nb_series_per_chunk=500)
#'
#' # Same example, from binary file
#' bin_file = "/tmp/epclust_series.bin"
#' nbytes = 8
#' endian = "little"
#' epclust::binarize(csv_file, bin_file, 500, nbytes, endian)
#' getSeries = function(indices) getDataInFile(indices, bin_file, nbytes, endian)
#' medoids_bin = claws(getSeries, K1=60, K2=6, "d8", "rel", nb_series_per_chunk=500)
#' unlink(csv_file)
#' unlink(bin_file)
#'
#' # Same example, from SQLite database
#' library(DBI)
#' series_db <- dbConnect(RSQLite::SQLite(), "file::memory:")
#' # Prepare data.frame in DB-format
#' n = nrow(series)
#' time_values = data.frame(
#'   id = rep(1:n,each=L),
#'   time = rep( as.POSIXct(1800*(0:n),"GMT",origin="2001-01-01"), L ),
#'   value = as.double(t(series)) )
#' dbWriteTable(series_db, "times_values", times_values)
#' # Fill associative array, map index to identifier
#' indexToID_inDB <- as.character(
#'   dbGetQuery(series_db, 'SELECT DISTINCT id FROM time_values')[,"id"] )
#' getSeries = function(indices) {
#'   request = "SELECT id,value FROM times_values WHERE id in ("
#'   for (i in indices)
#'     request = paste(request, i, ",", sep="")
#'   request = paste(request, ")", sep="")
#'   df_series = dbGetQuery(series_db, request)
#'   # Assume that all series share same length at this stage
#'   ts_length = sum(df_series[,"id"] == df_series[1,"id"])
#'   t( as.matrix(df_series[,"value"], nrow=ts_length) )
#' }
#' medoids_db = claws(getSeries, K1=60, K2=6, "d8", "rel", nb_series_per_chunk=500)
#' dbDisconnect(series_db)
#'
#' # All computed medoids should be the same:
#' digest::sha1(medoids_ascii)
#' digest::sha1(medoids_csv)
#' digest::sha1(medoids_bin)
#' digest::sha1(medoids_db)
#' }
#' @export
claws = function(getSeries, K1, K2,
	wf,ctype, #stage 1
	WER="end", #stage 2
	random=TRUE, #randomize series order?
	ntasks=1, ncores_tasks=1, ncores_clust=4, #control parallelism
	nb_series_per_chunk=50*K1, min_series_per_chunk=5*K1, #chunk size
	sep=",", #ASCII input separator
	nbytes=4, endian=.Platform$endian, #serialization (write,read)
	verbose=FALSE)
{
	# Check/transform arguments
	if (!is.matrix(getSeries) && !is.function(getSeries) &&
		!methods::is(getSeries, "connection" && !is.character(getSeries)))
	{
		stop("'getSeries': matrix, function, file or valid connection (no NA)")
	}
	K1 = .toInteger(K1, function(x) x>=2)
	K2 = .toInteger(K2, function(x) x>=2)
	if (!is.logical(random))
		stop("'random': logical")
	tryCatch(
		{ignored <- wavelets::wt.filter(wf)},
		error = function(e) stop("Invalid wavelet filter; see ?wavelets::wt.filter"))
	if (WER!="end" && WER!="mix")
		stop("WER takes values in {'end','mix'}")
	ntasks = .toInteger(ntasks, function(x) x>=1)
	ncores_tasks = .toInteger(ncores_tasks, function(x) x>=1)
	ncores_clust = .toInteger(ncores_clust, function(x) x>=1)
	nb_series_per_chunk = .toInteger(nb_series_per_chunk, function(x) x>=K1)
	min_series_per_chunk = .toInteger(K1, function(x) x>=K1 && x<=nb_series_per_chunk)
	if (!is.character(sep))
		stop("'sep': character")
	nbytes = .toInteger(nbytes, function(x) x==4 || x==8)

	# Serialize series if required, to always use a function
	bin_dir = ".epclust_bin/"
	dir.create(bin_dir, showWarnings=FALSE, mode="0755")
	if (!is.function(getSeries))
	{
		if (verbose)
			cat("...Serialize time-series\n")
		series_file = paste(bin_dir,"data",sep="") ; unlink(series_file)
		binarize(getSeries, series_file, nb_series_per_chunk, sep, nbytes, endian)
		getSeries = function(inds) getDataInFile(inds, series_file, nbytes, endian)
	}

	# Serialize all computed wavelets contributions onto a file
	contribs_file = paste(bin_dir,"contribs",sep="") ; unlink(contribs_file)
	index = 1
	nb_curves = 0
	if (verbose)
		cat("...Compute contributions and serialize them\n")
	repeat
	{
		series = getSeries((index-1)+seq_len(nb_series_per_chunk))
		if (is.null(series))
			break
		contribs_chunk = curvesToContribs(series, wf, ctype)
		binarize(contribs_chunk, contribs_file, nb_series_per_chunk, sep, nbytes, endian)
		index = index + nb_series_per_chunk
		nb_curves = nb_curves + nrow(contribs_chunk)
	}
	getContribs = function(indices) getDataInFile(indices, contribs_file, nbytes, endian)

	if (nb_curves < min_series_per_chunk)
		stop("Not enough data: less rows than min_series_per_chunk!")
	nb_series_per_task = round(nb_curves / ntasks)
	if (nb_series_per_task < min_series_per_chunk)
		stop("Too many tasks: less series in one task than min_series_per_chunk!")

	# Cluster contributions in parallel (by nb_series_per_chunk)
	indices_all = if (random) sample(nb_curves) else seq_len(nb_curves)
	indices_tasks = lapply(seq_len(ntasks), function(i) {
		upper_bound = ifelse( i<ntasks, min(nb_series_per_task*i,nb_curves), nb_curves )
		indices_all[((i-1)*nb_series_per_task+1):upper_bound]
	})
	if (verbose)
		cat(paste("...Run ",ntasks," x stage 1 in parallel\n",sep=""))
#	cl = parallel::makeCluster(ncores_tasks)
#	parallel::clusterExport(cl, varlist=c("getSeries","getContribs","K1","K2",
#		"nb_series_per_chunk","ncores_clust","synchrones_file","sep","nbytes","endian"),
#		envir = environment())
	# 1000*K1 indices [if WER=="end"], or empty vector [if WER=="mix"] --> series on file
#	indices = unlist( parallel::parLapply(cl, indices_tasks, function(inds) {
	indices = unlist( lapply(indices_tasks, function(inds) {
#		require("epclust", quietly=TRUE)

		browser() #TODO: CONTINUE DEBUG HERE

		indices_medoids = clusteringTask(inds,getContribs,K1,nb_series_per_chunk,ncores_clust)
		if (WER=="mix")
		{
			medoids2 = computeClusters2(
				getSeries(indices_medoids), K2, getSeries, nb_series_per_chunk)
			binarize(medoids2, synchrones_file, nb_series_per_chunk, sep, nbytes, endian)
			return (vector("integer",0))
		}
		indices_medoids
	}) )
#	parallel::stopCluster(cl)

	getRefSeries = getSeries
	synchrones_file = paste(bin_dir,"synchrones",sep="") ; unlink(synchrones_file)
	if (WER=="mix")
	{
		indices = seq_len(ntasks*K2)
		#Now series must be retrieved from synchrones_file
		getSeries = function(inds) getDataInFile(inds, synchrones_file, nbytes, endian)
		#Contributions must be re-computed
		unlink(contribs_file)
		index = 1
		if (verbose)
			cat("...Serialize contributions computed on synchrones\n")
		repeat
		{
			series = getSeries((index-1)+seq_len(nb_series_per_chunk))
			if (is.null(series))
				break
			contribs_chunk = curvesToContribs(series, wf, ctype)
			binarize(contribs_chunk, contribs_file, nb_series_per_chunk, sep, nbytes, endian)
			index = index + nb_series_per_chunk
		}
	}

	# Run step2 on resulting indices or series (from file)
	if (verbose)
		cat("...Run final // stage 1 + stage 2\n")
	indices_medoids = clusteringTask(
		indices, getContribs, K1, nb_series_per_chunk, ncores_tasks*ncores_clust)
	medoids = computeClusters2(getSeries(indices_medoids),K2,getRefSeries,nb_series_per_chunk)

	# Cleanup
	unlink(bin_dir, recursive=TRUE)

	medoids
}

#' curvesToContribs
#'
#' Compute the discrete wavelet coefficients for each series, and aggregate them in
#' energy contribution across scales as described in https://arxiv.org/abs/1101.4744v2
#'
#' @param series Matrix of series (in rows), of size n x L
#' @inheritParams claws
#'
#' @return A matrix of size n x log(L) containing contributions in rows
#'
#' @export
curvesToContribs = function(series, wf, ctype)
{
	L = length(series[1,])
	D = ceiling( log2(L) )
	nb_sample_points = 2^D
	cont_types = c("relative","absolute")
	ctype = cont_types[ pmatch(ctype,cont_types) ]
	t( apply(series, 1, function(x) {
		interpolated_curve = spline(1:L, x, n=nb_sample_points)$y
		W = wavelets::dwt(interpolated_curve, filter=wf, D)@W
		nrj = rev( sapply( W, function(v) ( sqrt( sum(v^2) ) ) ) )
		if (ctype=="relative") nrj / sum(nrj) else nrj
	}) )
}

# Helper for main function: check integer arguments with functiional conditions
.toInteger <- function(x, condition)
{
	if (!is.integer(x))
		tryCatch(
			{x = as.integer(x)[1]},
			error = function(e) paste("Cannot convert argument",substitute(x),"to integer")
		)
	if (!condition(x))
		stop(paste("Argument",substitute(x),"does not verify condition",body(condition)))
	x
}
